from .maml import MAML
from .classifier import Classifier
from .diverse_maml import DiverseMAML
from .diverse_transfer import DiverseTransfer
from .few_shot_adapter import FewShotAdapter
from .cross_domain_classifier import CrossDomainClassifier



import logging
from collections import OrderedDict
from functools import partial
from typing import Callable

import torch
import torchmetrics
from omegaconf import DictConfig

from torchcross.cd.heads import new_classification_head
from torchcross.cd.metrics import get_accuracy_func, get_auroc_func
from torchcross.cd.losses import get_loss_func
from torchcross.data.task import TaskTarget, TaskDescription
from torchcross.models.lightning import Classifier


class DiverseTransfer(Classifier):
    def __init__(
        self,
        backbone: tuple[torch.nn.Module, int],
        optim_cfg,
        hparams: DictConfig,
        *args,
        **kwargs,
    ) -> None:
        super(Classifier, self).__init__(*args, **kwargs)

        self.cfg = hparams
        self.save_hyperparameters(hparams)

        self.backbone, self.num_backbone_features = backbone

        self.optim_cfg = optim_cfg

        self.train_accuracy = torchmetrics.MeanMetric()
        self.train_AUROC = torchmetrics.MeanMetric()
        self.val_accuracy = torchmetrics.MeanMetric()
        self.val_AUROC = torchmetrics.MeanMetric()
        self.test_accuracy = torchmetrics.MeanMetric()
        self.test_AUROC = torchmetrics.MeanMetric()

        self.heads = torch.nn.ModuleDict()

    def forward(
        self, x: torch.Tensor, task_description: TaskDescription
    ) -> torch.Tensor:
        head = self.get_head(task_description, create=True, device=x.device)
        net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))
        return net(x)

    def training_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        (x, y), batch_task_description = batch

        if x.shape[1] == 1:
            x = x.repeat(1, 3, 1, 1)

        head = self.get_head(batch_task_description, create=True, device=x.device)
        loss_func = self.get_loss_func(batch_task_description, device=y.device)
        pred_func = self.get_pred_func(batch_task_description)
        accuracy_func, auroc_func = self.get_metric_func(
            batch_task_description, device=y.device
        )
        net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))

        logits = net(x)
        with torch.no_grad():
            pred = pred_func(logits)
            self.train_accuracy(accuracy_func(pred, y))
            self.train_AUROC(auroc_func(pred, y))
        loss = loss_func(logits, y)

        self.log("loss/train", loss, batch_size=len(y))
        self.log_dict(
            {
                "accuracy/train": self.train_accuracy,
                "AUROC/train": self.train_AUROC,
            },
            prog_bar=True,
        )
        return loss

    def validation_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        (x, y), batch_task_description = batch

        if x.shape[1] == 1:
            x = x.repeat(1, 3, 1, 1)

        if len(self.heads) == 0:
            logging.info("No heads found. Skipping validation step.")
            return

        loss_func = self.get_loss_func(batch_task_description, device=y.device)
        pred_func = self.get_pred_func(batch_task_description)
        accuracy_func, auroc_func = self.get_metric_func(
            batch_task_description, device=y.device
        )

        logits = self.forward(x, batch_task_description)
        pred = pred_func(logits)
        self.val_accuracy(accuracy_func(pred, y))
        self.val_AUROC(auroc_func(pred, y))
        loss = loss_func(logits, y)

        self.log("loss/val", loss, batch_size=len(y))
        self.log_dict(
            {
                "accuracy/val": self.val_accuracy,
                "AUROC/val": self.val_AUROC,
            },
            prog_bar=True,
        )

    def test_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        (x, y), batch_task_description = batch

        if x.shape[1] == 1:
            x = x.repeat(1, 3, 1, 1)

        loss_func = self.get_loss_func(batch_task_description, device=y.device)
        pred_func = self.get_pred_func(batch_task_description)
        accuracy_func, auroc_func = self.get_metric_func(
            batch_task_description, device=y.device
        )

        logits = self.forward(x, batch_task_description)
        pred = pred_func(logits)
        self.test_accuracy(accuracy_func(pred, y))
        self.test_AUROC(auroc_func(pred, y))
        loss = loss_func(logits, y)

        self.log("loss/test", loss, batch_size=len(y))
        self.log_dict(
            {
                "accuracy/test": self.test_accuracy,
                "AUROC/test": self.test_AUROC,
            },
            prog_bar=True,
        )

    def get_head(
        self, task_description: TaskDescription, create=False, device=None
    ) -> torch.nn.Module:
        task_identifier = task_description.task_identifier
        if task_identifier not in self.heads:
            if create:
                if device is None:
                    raise ValueError("Device must be specified if creating head.")
                new_head = new_classification_head(
                    task_description.task_target,
                    task_description.classes,
                    self.num_backbone_features,
                    device,
                )
                self.heads[task_identifier] = new_head
                self.optimizers().add_param_group({"params": new_head.parameters()})
            else:
                raise ValueError(f"Head for task {task_identifier} not found.")
        return self.heads[task_identifier]

    def get_loss_func(
        self, task_description: TaskDescription, device=None
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        return get_loss_func(
            task_description.task_target, task_description.classes, device
        )

    def get_pred_func(
        self, task_description: TaskDescription
    ) -> Callable[[torch.Tensor], torch.Tensor]:
        match task_description:
            case TaskDescription(task_target=TaskTarget.MULTICLASS_CLASSIFICATION):
                return partial(torch.softmax, dim=-1)
            case TaskDescription(
                task_target=(
                    TaskTarget.MULTILABEL_CLASSIFICATION
                    | TaskTarget.BINARY_CLASSIFICATION
                )
            ):
                return torch.sigmoid
            case TaskDescription(task_target=(TaskTarget.ORDINAL_REGRESSION)):
                return partial(torch.softmax, dim=-1)
            case TaskDescription(task_target):
                raise NotImplementedError(
                    f"Task target {task_target} not yet implemented"
                )

    def get_metric_func(
        self, task_description: TaskDescription, device=None
    ) -> tuple[
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    ]:
        return (
            get_accuracy_func(
                task_description.task_target, task_description.classes, device
            ),
            get_auroc_func(
                task_description.task_target, task_description.classes, device
            ),
        )



import torch

from torchcross.data.task import TaskTarget


def new_classification_head(
    task_target: TaskTarget,
    classes: dict[int, str],
    num_in_features: int,
    device: torch.device,
) -> torch.nn.Module:
    match task_target:
        case (
            TaskTarget.MULTICLASS_CLASSIFICATION
            | TaskTarget.MULTILABEL_CLASSIFICATION
            | TaskTarget.ORDINAL_REGRESSION
        ):
            fc = torch.nn.Linear(num_in_features, len(classes), device=device)
        case TaskTarget.BINARY_CLASSIFICATION | TaskTarget.REGRESSION:
            fc = torch.nn.Linear(num_in_features, 1, device=device)
        case target:
            raise NotImplementedError(f"Task target {target} is not supported")
    fc.weight.data = torch.zeros_like(fc.weight.data)
    fc.bias.data = torch.zeros_like(fc.bias.data)
    return fc



from typing import Any, Dict, Sequence, Tuple, Union

import hydra
import lightning.pytorch as pl
import torch
import torchmetrics
from omegaconf import DictConfig
from torch.optim import Optimizer
import torch.nn.functional as F


class Classifier(pl.LightningModule):
    def __init__(
        self,
        net: torch.nn.Module,
        optim_cfg,
        hparams: DictConfig,
        num_classes,
        *args,
        **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.save_hyperparameters(hparams)

        self.optim_cfg = optim_cfg
        self.net = net

        self.train_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.train_AUROC = torchmetrics.AUROC("multiclass", num_classes=num_classes)
        self.val_accuracy = torchmetrics.Accuracy("multiclass", num_classes=num_classes)
        self.val_AUROC = torchmetrics.AUROC("multiclass", num_classes=num_classes)
        self.test_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.test_AUROC = torchmetrics.AUROC("multiclass", num_classes=num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)

    def training_step(self, batch: Any, batch_idx: int):
        x, y = batch
        logits = self.forward(x)
        with torch.no_grad():
            pred = torch.softmax(logits, dim=-1)
            self.train_accuracy(pred, y)
            self.train_AUROC(pred, y)
        loss = F.cross_entropy(logits, y)

        self.log("loss/train", loss, batch_size=len(y))
        self.log_dict(
            {
                "accuracy/train": self.train_accuracy,
                "AUC/train": self.train_AUROC,
            },
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch: Any, batch_idx: int):
        x, y = batch
        logits = self.forward(x)
        pred = torch.softmax(logits, dim=-1)
        self.val_accuracy(pred, y)
        self.val_AUROC(pred, y)
        loss = F.cross_entropy(logits, y)

        self.log("loss/val", loss, batch_size=len(y))
        self.log_dict(
            {
                "accuracy/val": self.val_accuracy,
                "AUC/val": self.val_AUROC,
            },
            prog_bar=True,
        )

    def test_step(self, batch: Any, batch_idx: int):
        x, y = batch
        logits = self.forward(x)
        pred = torch.softmax(logits, dim=-1)
        self.test_accuracy(pred, y)
        self.test_AUROC(pred, y)
        loss = F.cross_entropy(logits, y)

        self.log("loss/test", loss, batch_size=len(y))
        self.log_dict(
            {
                "accuracy/test": self.test_accuracy,
                "AUC/test": self.test_AUROC,
            },
            prog_bar=True,
        )

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        x, y = batch
        logits = self.forward(x)
        pred = torch.softmax(logits, dim=-1)
        return pred

    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Tuple[Sequence[Optimizer], Sequence[Any]]]:
        outer_optimizer = self.optim_cfg.outer_optimizer(params=self.parameters())

        if self.optim_cfg.use_lr_scheduler:
            scheduler = hydra.utils.instantiate(
                self.optim_cfg.lr_scheduler, optimizer=outer_optimizer
            )
            return [outer_optimizer], [scheduler]

        return outer_optimizer



from collections import OrderedDict
from functools import partial
from typing import Sequence, Callable

import torch
import torch.nn.functional as F
import torchmetrics
import torchopt
from omegaconf import DictConfig
from torchmetrics.metric import Metric

from torchcross.cd.heads import new_classification_head
from torchcross.cd.losses import get_loss_func
from torchcross.cd.metrics import get_accuracy_func
from torchcross.data.task import Task, TaskTarget
from torchcross.models.lightning import MAML


class DiverseMAML(MAML):
    def __init__(
        self,
        backbone: tuple[torch.nn.Module, int],
        # head_init: Literal['kaiming', 'zero', 'adaptive', 'unicorn', 'proto'] = 'zero',
        optim_cfg,
        hparams: DictConfig,
        inductive=True,
        *args,
        **kwargs,
    ) -> None:
        super(MAML, self).__init__(*args, **kwargs)

        self.cfg = hparams
        self.save_hyperparameters(hparams)

        self.backbone, self.num_backbone_features = backbone

        self.optim_cfg = optim_cfg

        self.inductive = inductive

        self.automatic_optimization = False

        self.train_support_accuracy = torchmetrics.MeanMetric()
        self.train_query_accuracy = torchmetrics.MeanMetric()
        self.val_support_accuracy = torchmetrics.MeanMetric()
        self.val_query_accuracy = torchmetrics.MeanMetric()
        self.test_support_accuracy = torchmetrics.MeanMetric()
        self.test_query_accuracy = torchmetrics.MeanMetric()

    def outer_loop(
        self, batch: Sequence[Task], train: bool, metrics: dict[str, Metric]
    ):
        mode = "metatrain" if train else "metaval"

        support_losses = []
        query_losses = []
        support_accuracies = []
        query_accuracies = []

        backbone_state_dict = torchopt.extract_state_dict(
            self.backbone, by="reference", detach_buffers=True
        )

        for task in batch:
            support_x, support_y = task.support
            query_x, query_y = task.query

            if support_x.shape[1] == 1:
                support_x = support_x.repeat(1, 3, 1, 1)
                query_x = query_x.repeat(1, 3, 1, 1)

            head = self.get_head(task)
            loss_func = self.get_loss_func(task)
            pred_func = self.get_pred_func(task)
            metric_func = self.get_metric_func(task)
            net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))

            inner_optimizer = self.optim_cfg.inner_optimizer(net)
            optim_state_dict = torchopt.extract_state_dict(
                inner_optimizer, by="reference"
            )

            net.train()

            for k in range(self.cfg.num_inner_steps):
                support_logit = net(support_x)
                loss = loss_func(support_logit, support_y)
                inner_optimizer.step(loss)

            if self.inductive:
                net.eval()

            with torch.no_grad():
                support_logit = net(support_x)
                support_losses.append(loss_func(support_logit, support_y))
                support_pred = pred_func(support_logit)
                support_accuracies.append(metric_func(support_pred, support_y))

            query_logit = net(query_x)
            query_losses.append(loss_func(query_logit, query_y))
            with torch.no_grad():
                query_pred = pred_func(query_logit)
                query_accuracies.append(metric_func(query_pred, query_y))

            torchopt.recover_state_dict(self.backbone, backbone_state_dict)
            # TODO: find out if the following line is even needed since the optimizer is only used for this single task
            torchopt.recover_state_dict(inner_optimizer, optim_state_dict)

        support_loss = sum(support_losses) / len(support_losses)
        query_loss = sum(query_losses) / len(query_losses)

        support_accuracy = sum(support_accuracies) / len(support_accuracies)
        query_accuracy = sum(query_accuracies) / len(query_accuracies)

        self.log(f"loss/{mode}/support", support_loss, batch_size=len(batch))
        self.log(f"loss/{mode}/query", query_loss, batch_size=len(batch), prog_bar=True)

        metrics["support_accuracy"](support_accuracy)
        metrics["query_accuracy"](query_accuracy)

        return query_loss

    def get_head(self, task: Task) -> torch.nn.Module:
        return new_classification_head(
            task.task_target,
            task.classes,
            self.num_backbone_features,
            task.support[0].device,
        )

    def get_loss_func(
        self, task: Task
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        return get_loss_func(task.task_target, task.classes, task.support[0].device)

    def get_pred_func(self, task: Task) -> Callable[[torch.Tensor], torch.Tensor]:
        match task:
            case Task(task_target=TaskTarget.MULTICLASS_CLASSIFICATION):
                return partial(torch.softmax, dim=-1)
            case Task(
                task_target=(
                    TaskTarget.MULTILABEL_CLASSIFICATION
                    | TaskTarget.BINARY_CLASSIFICATION
                )
            ):
                return torch.sigmoid
            case Task(task_target=(TaskTarget.ORDINAL_REGRESSION)):
                return partial(torch.softmax, dim=-1)
            case Task(task_target):
                raise NotImplementedError(
                    f"Task target {task_target} not yet implemented"
                )

    def get_metric_func(
        self, task: Task
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        return get_accuracy_func(task.task_target, task.classes, task.support[0].device)



from .base import MetaDataset, MetaConcatDataset
from .few_shot import (
    FewShotMetaDataset,
    SubTaskRandomFewShotMetaDataset,
    SubtaskMetaDataset,
)
from .few_shot_dummies import TakeFirstFewShotMetaDataset



from collections import OrderedDict
from typing import Any, Sequence, Tuple, Union

import torchopt
import hydra
import lightning.pytorch as pl
import torch
import torchmetrics
from omegaconf import DictConfig
from torch.optim import Optimizer
import torch.nn.functional as F
from torchmetrics.metric import Metric

from torchcross.data.task import Task


class EpisodicModel(pl.LightningModule):
    def __init__(
        self,
        net: torch.nn.Module,
        optim_cfg,
        hparams: DictConfig,
        num_classes,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.cfg = hparams
        self.save_hyperparameters(hparams)

        self.net = net

        self.optim_cfg = optim_cfg

        self.automatic_optimization = False

        self.train_support_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.train_query_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.val_support_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.val_query_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.test_support_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.test_query_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )

    def forward(self, batch) -> torch.Tensor:

        backbone_state_dict = torchopt.extract_state_dict(
            self.backbone, by="reference", detach_buffers=True
        )

        query_logits = []

        for task in batch:
            support_x, support_y = task.support
            query_x, query_y = task.query

            if support_x.shape[1] == 1:
                support_x = support_x.repeat(1, 3, 1, 1)
                query_x = query_x.repeat(1, 3, 1, 1)

            head = self.get_head(task)
            loss_func = self.get_loss_func(task)
            pred_func = self.get_pred_func(task)
            metric_func = self.get_metric_func(task)
            net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))
            self.inner_loop(net, support_x, support_y, loss_func)
            query_logits.append(net(query_x))
            torchopt.recover_state_dict(self.backbone, backbone_state_dict)

    def outer_loop(
        self, batch: Sequence[Task], train: bool, metrics: dict[str, Metric]
    ):
        mode = "metatrain" if train else "metaval"

        support_losses = []
        query_losses = []
        support_accuracies = []
        query_accuracies = []

        backbone_state_dict = torchopt.extract_state_dict(
            self.backbone, by="reference", detach_buffers=True
        )

        for task in batch:
            sl, ql, sm, qm = self.episode(task)
            torchopt.recover_state_dict(self.backbone, backbone_state_dict)

        support_loss = sum(support_losses) / len(support_losses)
        query_loss = sum(query_losses) / len(query_losses)

        support_accuracy = sum(support_accuracies) / len(support_accuracies)
        query_accuracy = sum(query_accuracies) / len(query_accuracies)

        self.log(f"loss/{mode}/support", support_loss, batch_size=len(batch))
        self.log(f"loss/{mode}/query", query_loss, batch_size=len(batch), prog_bar=True)

        metrics["support_accuracy"](support_accuracy)
        metrics["query_accuracy"](query_accuracy)

        return support_loss, query_loss, support_metrics, query_metrics

    def episode(self, task):
        support_x, support_y = task.support
        query_x, query_y = task.query
        if support_x.shape[1] == 1:
            support_x = support_x.repeat(1, 3, 1, 1)
            query_x = query_x.repeat(1, 3, 1, 1)
        head = self.get_head(task)
        loss_func = self.get_loss_func(task)
        pred_func = self.get_pred_func(task)
        metric_func = self.get_metric_func(task)
        net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))
        self.inner_loop(net, support_x, support_y, loss_func)
        with torch.no_grad():
            support_logit = net(support_x)
            support_loss = loss_func(support_logit, support_y)
            support_pred = pred_func(support_logit)
            support_metric = metric_func(support_pred, support_y)
        query_logit = net(query_x)
        query_loss = loss_func(query_logit, query_y)
        with torch.no_grad():
            query_pred = pred_func(query_logit)
            query_metric = metric_func(query_pred, query_y)
        return support_loss, query_loss, support_metric, query_metric

    def inner_loop(self, net, support_x, support_y, loss_func):
        inner_optimizer = self.optim_cfg.inner_optimizer(net)
        net.train()
        for k in range(self.cfg.num_inner_steps):
            support_logit = net(support_x)
            loss = loss_func(support_logit, support_y)
            inner_optimizer.step(loss)

    def training_step(self, batch: Any, batch_idx: int):
        outer_optimizer = self.optimizers()
        outer_optimizer.zero_grad()

        loss = self.outer_loop(
            batch,
            True,
            {
                "support_accuracy": self.train_support_accuracy,
                "query_accuracy": self.train_query_accuracy,
            },
        )

        self.manual_backward(loss)

        self.log(
            "accuracy/metatrain/support_accuracy",
            self.train_support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metatrain/query_accuracy",
            self.train_query_accuracy,
            prog_bar=True,
        )

        self.optimizers().step()

    def validation_step(self, batch: Any, batch_idx: int):
        torch.set_grad_enabled(True)
        self.train()
        self.outer_loop(
            batch,
            False,
            {
                "support_accuracy": self.val_support_accuracy,
                "query_accuracy": self.val_query_accuracy,
            },
        )
        torch.set_grad_enabled(False)

        self.log(
            "accuracy/metaval/support_accuracy",
            self.val_support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metaval/query_accuracy",
            self.val_query_accuracy,
            prog_bar=True,
        )

    def test_step(self, batch: Any, batch_idx: int):
        torch.set_grad_enabled(True)
        self.train()
        self.outer_loop(
            batch,
            False,
            {
                "support_accuracy": self.test_support_accuracy,
                "query_accuracy": self.test_query_accuracy,
            },
        )
        torch.set_grad_enabled(False)

        self.log(
            "accuracy/metaval/support_accuracy",
            self.test_support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metaval/query_accuracy",
            self.test_query_accuracy,
            prog_bar=True,
        )

    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Tuple[Sequence[Optimizer], Sequence[Any]]]:

        outer_optimizer = self.optim_cfg.outer_optimizer(params=self.parameters())

        if self.optim_cfg.use_lr_scheduler:
            scheduler = hydra.utils.instantiate(
                self.optim_cfg.lr_scheduler, optimizer=outer_optimizer
            )
            return [outer_optimizer], [scheduler]

        return outer_optimizer



from typing import Callable

import torch
import torch.nn.functional as F

from torchcross.data.task import TaskTarget


def get_loss_func(
    task_target: TaskTarget, classes: dict[int, str], device: torch.device = None
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return F.cross_entropy
        case TaskTarget.MULTILABEL_CLASSIFICATION | TaskTarget.BINARY_CLASSIFICATION:
            return lambda p, y: F.binary_cross_entropy_with_logits(p, y.float())
        case TaskTarget.ORDINAL_REGRESSION:
            class_to_idx = torch.full(
                (max(classes) + 1,),
                -1,
                dtype=torch.long,
                device=device,
            )
            for i, c in enumerate(classes):
                class_to_idx[c] = i
            return lambda p, y: F.cross_entropy(p, class_to_idx[y])
        case TaskTarget.REGRESSION:
            return F.mse_loss
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")


def get_ordinal_cross_entropy_loss(classes):
    # Compute the weights for ordinal cross-entropy loss
    # using the intervals between consecutive class keys.
    sorted_class_keys = sorted(classes.keys())
    sorted_keys_tensor = torch.tensor(sorted_class_keys, dtype=torch.float)
    label_intervals_tensor = torch.diff(sorted_keys_tensor)
    weights = torch.cat((label_intervals_tensor, torch.tensor([1])), dim=0)

    def ordinal_cross_entropy(p, y):
        y = y.to(torch.long).unsqueeze(1)
        y_binary = y <= torch.arange(p.size(1), device=y.device)

        weightss = weights.unsqueeze(0).repeat(y.size(0), 1).to(p.device)

        binary_losses = F.binary_cross_entropy_with_logits(
            p, y_binary.float()
        )
        weighted_binary_losses = binary_losses * weightss
        ordinal_loss = weighted_binary_losses.mean()
        return ordinal_loss

    return ordinal_cross_entropy



import itertools
import random
from builtins import NotImplementedError
from collections.abc import Sequence, Iterable, Iterator, Callable
from functools import partial

import numpy as np
import torch
from torch.utils.data import SubsetRandomSampler

from torchcross.utils.collate_fn import identity
from . import MetaDataset
from .base import IterableMetaDataset
from ..base import TaskSource
from ..task import TaskTarget, Task


def get_indices(labels, classes, task_target):
    if isinstance(labels, np.ndarray):

        def ind_func(x):
            return np.nonzero(x)[0]  # .tolist()

    elif isinstance(labels, torch.Tensor):

        def ind_func(x):
            return torch.nonzero(x).flatten()  # .tolist()

    else:
        raise ValueError("Unsupported label data type")

    if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
        pos_indices = {c: ind_func(labels == c) for c in classes}
        neg_indices = {c: ind_func(labels != c) for c in classes}
    elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
        pos_indices = {c: ind_func(labels[:, c]) for c in classes}
        neg_indices = {c: ind_func(labels[:, c] - 1) for c in classes}
    elif task_target is TaskTarget.BINARY_CLASSIFICATION:
        pos_indices = {1: ind_func(labels)}
        neg_indices = {1: ind_func(labels - 1)}
    else:
        raise NotImplementedError(f"Task target {task_target} not yet implemented")
    return pos_indices, neg_indices


def multiclass_few_shot_sample(
    pos_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    n_shot: int,
    selected_classes: Sequence[int] | dict[int, str],
):
    sample = np.array([], dtype=np.int64)
    # Sample n_shot positive examples for each class
    for c in selected_classes:
        pos_sample = np.random.choice(pos_indices[c], size=n_shot, replace=False)
        sample = np.concatenate([sample, pos_sample])
    return sample


def binary_few_shot_sample(
    pos_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    neg_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    n_shot: int,
):
    pos_sample = np.random.choice(pos_indices[1], size=n_shot, replace=False)
    neg_sample = np.random.choice(neg_indices[1], size=n_shot, replace=False)
    return np.concatenate([pos_sample, neg_sample])


def multilabel_few_shot_sample(
    labels: np.ndarray | torch.Tensor,
    pos_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    neg_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    n_shot: int,
    selected_classes: Sequence[int] | dict[int, str],
) -> np.ndarray:
    if isinstance(labels, np.ndarray):
        sum_func = partial(np.sum, axis=0)
        all_func = np.all
    elif isinstance(labels, torch.Tensor):
        sum_func = partial(torch.sum, dim=0)
        all_func = torch.all
    else:
        raise ValueError("Unsupported data type")

    if isinstance(selected_classes, dict):
        selected_classes = list(selected_classes.keys())
    labels = labels[:, selected_classes]

    def criterion(sample_to_test: np.ndarray) -> bool:
        column_sums = sum_func(labels[sample_to_test])
        pos_criterion = all_func(column_sums >= n_shot)
        neg_criterion = all_func(column_sums <= len(sample_to_test) - n_shot)
        return pos_criterion and neg_criterion

    sample = np.array([], dtype=np.int64)

    # Sample n_shot positive and n_shot negative examples for each label
    for c in selected_classes:
        pos_sample = np.random.choice(pos_indices[c], size=n_shot, replace=False)
        neg_sample = np.random.choice(neg_indices[c], size=n_shot, replace=False)
        sample = np.concatenate([sample, pos_sample, neg_sample])

    # Remove duplicates and shuffle the sample for the later steps
    sample = np.unique(sample)
    np.random.shuffle(sample)

    # Add single examples from the sample until criterion is satisfied, i.e. until
    # there are enough positive and negative examples for each label.
    # We do this so we have to remove fewer examples (and therefore need to compute the
    # criterion fewer times) later on.
    for i in range(2 * n_shot, len(sample)):
        small_sample = sample[:i]
        if criterion(small_sample):
            sample = small_sample
            break

    # Remove examples (greedily) as long as criterion is satisfied without the example,
    # i.e. as long as there are enough positive and negative examples left for each
    # label.
    i = 0
    while i < len(sample):
        row_removed = np.delete(sample, i)
        if criterion(row_removed):
            sample = row_removed
        else:
            i += 1

    return sample


def remove_indices(
    pos_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    neg_indices: dict[int, np.ndarray | torch.Tensor | Sequence[int]],
    sample: np.ndarray | torch.Tensor | Sequence[int],
):
    sample_set = set(sample)

    def _compute(class_indices):
        if isinstance(class_indices, np.ndarray):
            return np.setdiff1d(class_indices, sample)
        elif isinstance(class_indices, torch.Tensor):
            return torch.from_numpy(np.setdiff1d(class_indices, sample))
            # return class_indices[
            #     torch.isin(class_indices, sample, assume_unique=True, invert=True)
            # ]
        else:
            return [i for i in class_indices if i not in sample_set]

    new_pos_indices = {c: _compute(c_indices) for c, c_indices in pos_indices.items()}
    new_neg_indices = {c: _compute(c_indices) for c, c_indices in neg_indices.items()}
    return new_pos_indices, new_neg_indices


class FewShotMetaDataset(IterableMetaDataset):
    def __init__(
        self,
        task_source: TaskSource,
        collate_fn=None,
        n_support_samples_per_class: int = 5,
        n_query_samples_per_class: int = 20,
        filter_classes_min_samples: int = 50,
        length: int | None = None,
        output_indices: bool = False,
    ) -> None:
        super().__init__()
        self.task_source = task_source
        self.task_target = task_source.task_target

        self.pos_indices, self.neg_indices = get_indices(
            self.task_source.labels, self.task_source.classes, self.task_target
        )

        self.used_classes = list(self.task_source.classes.keys())
        if self.task_target is TaskTarget.BINARY_CLASSIFICATION:
            self.used_classes = [1]
        if filter_classes_min_samples > 0:
            self.used_classes = [
                c
                for c in self.used_classes
                if len(self.pos_indices[c]) >= filter_classes_min_samples
            ]

            if self.task_target in (
                TaskTarget.BINARY_CLASSIFICATION,
                TaskTarget.MULTILABEL_CLASSIFICATION,
            ):
                self.used_classes = [
                    c
                    for c in self.used_classes
                    if len(self.neg_indices[c]) >= filter_classes_min_samples
                ]
        if (
            len(self.used_classes) < 2
            and self.task_target is not TaskTarget.BINARY_CLASSIFICATION
        ):
            message = "FewShotMetaDataset requires at least 2 classes"
            if filter_classes_min_samples > 0:
                message += f" with at least {filter_classes_min_samples} samples each"
            message += f"for {self.task_target}."
            raise ValueError(message)

        self.collate_fn = collate_fn if collate_fn else identity
        self.n_support_samples_per_class = n_support_samples_per_class
        self.n_query_samples_per_class = n_query_samples_per_class
        self.length = length
        self.output_indices = output_indices

    def __len__(self) -> int:
        if self.length is None:
            raise ValueError("Length is not set.")
        return self.length

    def __iter__(self) -> Iterator[Task]:
        match self.task_target:
            case TaskTarget.MULTICLASS_CLASSIFICATION | TaskTarget.ORDINAL_REGRESSION:
                return self._multiclass_iter()
            case TaskTarget.MULTILABEL_CLASSIFICATION:
                return self._multilabel_iter()
            case TaskTarget.BINARY_CLASSIFICATION:
                return self._binary_iter()
            case target:
                msg = f"Task target {target} is not supported by FewShotMetaDataset."
                raise NotImplementedError(msg)

    def _multiclass_iter(self) -> Iterator[Task]:
        pos_indices = self.pos_indices
        neg_indices = self.neg_indices
        iter_index = 0
        stop = False
        while not stop:
            try:
                support_indices = multiclass_few_shot_sample(
                    pos_indices, self.n_support_samples_per_class, self.used_classes
                )
                new_pos_indices, new_neg_indices = remove_indices(
                    pos_indices, neg_indices, support_indices
                )
                query_indices = multiclass_few_shot_sample(
                    new_pos_indices, self.n_query_samples_per_class, self.used_classes
                )
                if self.length is None:
                    pos_indices, neg_indices = remove_indices(
                        new_pos_indices, new_neg_indices, query_indices
                    )
                else:
                    iter_index += 1
                    if iter_index >= self.length:
                        stop = True
                class_to_label = {
                    c: torch.tensor(ci) for ci, c in enumerate(self.used_classes)
                }
                support = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        class_to_label[self.task_source.labels[i].item()],
                    )
                    for i in support_indices
                ]
                query = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        class_to_label[self.task_source.labels[i].item()],
                    )
                    for i in query_indices
                ]
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    ci: self.task_source.classes[c]
                    for ci, c in enumerate(self.used_classes)
                }
                yield Task(support, query, self.task_target, class_dict)
            except ValueError:
                return

    def _multilabel_iter(self) -> Iterator[Task]:
        pos_indices = self.pos_indices
        neg_indices = self.neg_indices
        iter_index = 0
        stop = False
        while not stop:
            try:
                support_indices = multilabel_few_shot_sample(
                    self.task_source.labels,
                    pos_indices,
                    neg_indices,
                    self.n_support_samples_per_class,
                    self.used_classes,
                )
                new_pos_indices, new_neg_indices = remove_indices(
                    pos_indices, neg_indices, support_indices
                )
                query_indices = multilabel_few_shot_sample(
                    self.task_source.labels,
                    new_pos_indices,
                    new_neg_indices,
                    self.n_query_samples_per_class,
                    self.used_classes,
                )
                if self.length is None:
                    pos_indices, neg_indices = remove_indices(
                        new_pos_indices, new_neg_indices, query_indices
                    )
                else:
                    iter_index += 1
                    if iter_index >= self.length:
                        stop = True

                labels = self.task_source.labels[:, self.used_classes]
                support = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in support_indices
                ]
                query = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in query_indices
                ]
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    ci: self.task_source.classes[c]
                    for ci, c in enumerate(self.used_classes)
                }
                yield Task(support, query, self.task_target, class_dict)
            except ValueError:
                return

    def _binary_iter(self) -> Iterator[Task]:
        pos_indices = self.pos_indices
        neg_indices = self.neg_indices
        iter_index = 0
        stop = False
        while not stop:
            try:
                support_indices = binary_few_shot_sample(
                    pos_indices, neg_indices, self.n_support_samples_per_class
                )
                new_pos_indices, new_neg_indices = remove_indices(
                    pos_indices, neg_indices, support_indices
                )
                query_indices = binary_few_shot_sample(
                    new_pos_indices, new_neg_indices, self.n_query_samples_per_class
                )
                if self.length is None:
                    pos_indices, neg_indices = remove_indices(
                        new_pos_indices, new_neg_indices, query_indices
                    )
                else:
                    iter_index += 1
                    if iter_index >= self.length:
                        stop = True

                labels = self.task_source.labels
                support = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in support_indices
                ]
                query = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in query_indices
                ]
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    0: self.task_source.classes[0],
                    1: self.task_source.classes[1],
                }
                yield Task(support, query, self.task_target, class_dict)

            except ValueError:
                return


class SubTaskRandomFewShotMetaDataset(FewShotMetaDataset):
    def __init__(
        self,
        task_source: TaskSource,
        collate_fn=None,
        *,
        n_support_samples_per_class_min: int = 1,
        n_support_samples_per_class_max: int = 10,
        n_query_samples_per_class: int = 10,
        filter_classes_min_samples: int = 40,
        length: int | None = None,
        output_indices: bool = False,
    ) -> None:
        super().__init__(
            task_source,
            collate_fn,
            n_support_samples_per_class_min,
            n_query_samples_per_class,
            filter_classes_min_samples,
            length,
            output_indices,
        )
        self.original_used_classes = self.used_classes
        self.n_support_samples_per_class_min = n_support_samples_per_class_min
        self.n_support_samples_per_class_max = n_support_samples_per_class_max

    def __iter__(self):
        if self.task_target != TaskTarget.BINARY_CLASSIFICATION:
            n_way = random.randint(2, len(self.original_used_classes))
            self.used_classes = random.sample(self.original_used_classes, n_way)
        self.n_support_samples_per_class = random.randint(
            self.n_support_samples_per_class_min, self.n_support_samples_per_class_max
        )
        for task in super().__iter__():
            yield task
            if self.task_target != TaskTarget.BINARY_CLASSIFICATION:
                n_way = random.randint(2, len(self.original_used_classes))
                self.used_classes = random.sample(self.original_used_classes, n_way)
            self.n_support_samples_per_class = random.randint(
                self.n_support_samples_per_class_min,
                self.n_support_samples_per_class_max,
            )
        self.used_classes = self.original_used_classes


class MultiLabelFewShotMetaDataset(IterableMetaDataset):
    def __init__(
        self,
        dataset: TaskSource,
        collate_fn=None,
        sampler: Callable[[Sequence[int]], Iterable[int]] = SubsetRandomSampler,
        n_support_samples_per_class: int = 5,
        n_query_samples_per_class: int = 20,
        filter_classes_min_samples: int = 50,
    ) -> None:
        super().__init__()
        self.dataset = dataset
        self.task_target = dataset.task_target

        self.pos_indices, self.neg_indices = get_indices(
            self.dataset.labels, self.dataset.classes, self.task_target
        )

        self.used_classes = list(self.dataset.classes.keys())
        if self.task_target is TaskTarget.BINARY_CLASSIFICATION:
            self.used_classes = [1]
        if filter_classes_min_samples > 0:
            self.used_classes = [
                c
                for c in self.used_classes
                if len(self.pos_indices[c]) >= filter_classes_min_samples
            ]

            if self.task_target in (
                TaskTarget.BINARY_CLASSIFICATION,
                TaskTarget.MULTILABEL_CLASSIFICATION,
            ):
                self.used_classes = [
                    c
                    for c in self.used_classes
                    if len(self.neg_indices[c]) >= filter_classes_min_samples
                ]
        if (
            len(self.used_classes) < 2
            and self.task_target is not TaskTarget.BINARY_CLASSIFICATION
        ):
            message = "FewShotMetaDataset requires at least 2 classes"
            if filter_classes_min_samples > 0:
                message += f" with at least {filter_classes_min_samples} samples each"
            message += f"for {self.task_target}."
            raise ValueError(message)

        self.collate_fn = collate_fn if collate_fn else identity
        self.samplers = {c: sampler(self.pos_indices[c]) for c in self.used_classes}
        self.n_support_samples_per_class = n_support_samples_per_class
        self.n_query_samples_per_class = n_query_samples_per_class

    def __len__(self) -> int:
        return min(len(s) for s in self.samplers.values()) // (
            self.n_support_samples_per_class + self.n_query_samples_per_class
        )

    def __iter__(self) -> Iterator[Task]:
        sampler_iters = [iter(s) for c, s in self.samplers.items()]

        def _get_multiclass_shot(n_shot_per_class):
            sample = [
                (i, j)
                for j, s in enumerate(sampler_iters)
                for i in [next(s) for _ in range(n_shot_per_class)]
            ]
            return [(self.dataset[i][0], torch.tensor(j)) for i, j in sample]

        while True:
            try:
                support = _get_multiclass_shot(self.n_support_samples_per_class)
                query = _get_multiclass_shot(self.n_query_samples_per_class)
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    ci: self.dataset.classes[c]
                    for ci, c in enumerate(self.used_classes)
                }
                yield Task(support, query, self.task_target, class_dict)
            except StopIteration:
                break


class SubtaskMetaDataset(MetaDataset):
    def __init__(
        self,
        dataset: TaskSource,
        collate_fn=None,
        n_shot_provider: Iterator[Sequence[int]] = None,
    ) -> None:
        super().__init__()
        self.dataset = dataset
        self.task_target = dataset.task_target

        self.indices = None

        if self.task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
            self.indices = {
                c: [i for i, l in enumerate(self.dataset.labels) if l == c]
                for c in self.dataset.classes
            }
        elif self.task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
            self.indices = {
                c: [i for i, l in enumerate(self.dataset.labels) if l[c]]
                for c in self.dataset.classes
            }
            self.anti_indices = {
                c: [i for i, l in enumerate(self.dataset.labels) if not l[c]]
                for c in self.dataset.classes
            }
        else:
            raise NotImplementedError(
                f"Task target {self.task_target} not yet implemented"
            )

        self.collate_fn = collate_fn if collate_fn else identity

        self.n_shot_provider = (
            n_shot_provider if n_shot_provider else itertools.repeat((5, None))
        )

    def sample_multiclass_shot(
        self,
        selected_classes: Sequence,
        n_shot_per_class: int | None = None,
        n_shot_total: int | None = None,
    ):
        if n_shot_per_class is None:
            if n_shot_total is None:
                raise ValueError(
                    "n_shot_per_class and n_shot_total can not both be None."
                )
            population = [
                (i, j) for j, c in enumerate(selected_classes) for i in self.indices[c]
            ]
            sample = random.sample(population, n_shot_total)
        else:
            if n_shot_total is not None:
                raise ValueError(
                    "n_shot_per_class and n_shot_total can not both be set."
                )
            sample = [
                (i, j)
                for j, c in enumerate(selected_classes)
                for i in random.sample(self.indices[c], n_shot_per_class)
            ]
        return [(self.dataset[i][0], torch.tensor(j)) for i, j in sample]

    def sample_multilabel_shot(
        self,
        selected_classes: Sequence,
        n_shot_per_class: int | None = None,
        n_shot_total: int | None = None,
    ):
        if n_shot_per_class is None:
            if n_shot_total is None:
                raise ValueError(
                    "n_shot_per_class and n_shot_total can not both be None."
                )

            criteria_fulfilled = False
            while not criteria_fulfilled:
                candidate = random.sample(range(len(self.dataset.labels)), n_shot_total)
                candidate_labels = self.dataset.labels[candidate]
                label_sums = sum(candidate_labels)
                criteria_fulfilled = all(0 < x < n_shot_total for x in label_sums)

            sample = [
                (i, [self.dataset.labels[i][c] for c in selected_classes])
                for i in candidate
            ]
        else:
            if n_shot_total is not None:
                raise ValueError(
                    "n_shot_per_class and n_shot_total can not both be set."
                )

            usable_indices = set(range(len(self.dataset.labels)))
            positive_sums = np.zeros(len(self.dataset.classes))
            negative_sums = np.zeros(len(self.dataset.classes))
            candidate = []
            criteria_fulfilled = False
            while not criteria_fulfilled:
                i = random.sample(usable_indices, 1)
                positive_sums = positive_sums + self.dataset.labels[i]
                negative_sums = negative_sums + 1 - self.dataset.labels[i]
                for c in selected_classes:
                    if positive_sums[c] == n_shot_per_class:
                        usable_indices.difference_update(self.indices[c])
                    if negative_sums[c] == n_shot_per_class:
                        usable_indices.difference_update(self.anti_indices[c])
                usable_indices.remove(i)
                candidate = candidate + i

            sample = [
                (i, [self.dataset.labels[i][c] for c in selected_classes])
                for i in candidate
            ]

        return [(self.dataset[i][0], torch.tensor(j)) for i, j in sample]

    def __getitem__(self, index: Sequence) -> Task:
        if self.task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
            support = self.collate_fn(
                self.sample_multiclass_shot(index, *next(self.n_shot_provider))
            )
            query = self.collate_fn(
                self.sample_multiclass_shot(index, *next(self.n_shot_provider))
            )
        elif self.task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
            support = self.collate_fn(
                self.sample_multilabel_shot(index, *next(self.n_shot_provider))
            )
            query = self.collate_fn(
                self.sample_multilabel_shot(index, *next(self.n_shot_provider))
            )
        else:
            raise NotImplementedError(
                f"Task target {self.task_target} not yet implemented"
            )

        return Task(
            support,
            query,
            self.task_target,
            {k: self.dataset.classes[k] for k in index},
        )



from collections import OrderedDict
from functools import partial
from typing import Sequence, Callable

import torch
import torch.nn.functional as F
import torchmetrics
import torchopt
import lightning.pytorch as pl
from omegaconf import DictConfig

from torchcross.cd.heads import new_classification_head
from torchcross.cd.losses import get_loss_func
from torchcross.cd.metrics import get_accuracy_func, get_auroc_func
from torchcross.data.task import Task, TaskTarget


class FewShotAdapter(pl.LightningModule):
    def __init__(
        self,
        backbone: tuple[torch.nn.Module, int],
        # head_init: Literal['kaiming', 'zero', 'adaptive', 'unicorn', 'proto'] = 'zero',
        optim_cfg,
        hparams: DictConfig,
        inductive=True,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.cfg = hparams
        self.save_hyperparameters(hparams)

        self.backbone, self.num_backbone_features = backbone

        self.optim_cfg = optim_cfg

        self.inductive = inductive

        self.automatic_optimization = False

        self.support_accuracy = torchmetrics.MeanMetric()
        self.query_accuracy = torchmetrics.MeanMetric()
        self.support_accuracy_list = torchmetrics.CatMetric()
        self.query_accuracy_list = torchmetrics.CatMetric()

        self.support_auroc = torchmetrics.MeanMetric()
        self.query_auroc = torchmetrics.MeanMetric()
        self.support_auroc_list = torchmetrics.CatMetric()
        self.query_auroc_list = torchmetrics.CatMetric()

        self.logged_values = []

    def test_step(self, batch: list[Task], batch_idx: int):
        torch.set_grad_enabled(True)
        self.train()
        self.outer_loop(batch)
        torch.set_grad_enabled(False)

        self.log(
            "accuracy/metatest/support",
            self.support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metatest/query",
            self.query_accuracy,
            prog_bar=False,
        )
        self.log(
            "auroc/metatest/support",
            self.support_auroc,
            prog_bar=False,
        )
        self.log(
            "auroc/metatest/query",
            self.query_auroc,
            prog_bar=False,
        )

    # def test_epoch_end(self, outputs: Union[EPOCH_OUTPUT, list[EPOCH_OUTPUT]]) -> None:
    # self.logger.experiment.add_text(
    #     "accuracy/metatest/support_list",
    #     str(self.support_accuracy_list.compute()),
    #     global_step=self.global_step,
    # )
    # self.logger.experiment.add_text(
    #     "accuracy/metatest/query_list",
    #     str(self.query_accuracy_list.compute()),
    #     global_step=self.global_step,
    # )
    # self.logger.experiment.add_text(
    #     "auroc/metatest/support_list",
    #     str(self.support_auroc_list.compute()),
    #     global_step=self.global_step,
    # )
    # self.logger.experiment.add_text(
    #     "auroc/metatest/query_list",
    #     str(self.query_auroc_list.compute()),
    #     global_step=self.global_step,
    # )

    # self.log(
    #     "accuracy/metatest/support_list",
    #     self.support_accuracy_list,
    #     prog_bar=False,
    # )
    # self.log(
    #     "accuracy/metatest/query_list",
    #     self.query_accuracy_list,
    #     prog_bar=False,
    # )
    # self.log(
    #     "auroc/metatest/support_list",
    #     self.support_auroc_list,
    #     prog_bar=False,
    # )
    # self.log(
    #     "auroc/metatest/query_list",
    #     self.query_auroc_list,
    #     prog_bar=False,
    # )

    def outer_loop(self, batch: Sequence[Task]):
        support_losses = []
        query_losses = []

        backbone_state_dict = torchopt.extract_state_dict(
            self.backbone, by="reference", detach_buffers=True
        )

        for task in batch:
            support_x, support_y = task.support
            query_x, query_y = task.query

            if support_x.shape[1] == 1:
                support_x = support_x.repeat(1, 3, 1, 1)
                query_x = query_x.repeat(1, 3, 1, 1)

            head = self.get_head(task)
            loss_func = self.get_loss_func(task)
            pred_func = self.get_pred_func(task)
            accuracy_func, auroc_func = self.get_metric_func(task)
            net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))

            inner_optimizer = self.optim_cfg.inner_optimizer(net.parameters())
            net.train()

            for k in range(self.cfg.num_inner_steps):
                support_logit = net(support_x)
                loss = loss_func(support_logit, support_y)
                inner_optimizer.zero_grad()  # zero gradients
                loss.backward()  # backward
                inner_optimizer.step()  # step updates

            if self.inductive:
                net.eval()

            with torch.no_grad():
                support_logit = net(support_x)
                support_losses.append(loss_func(support_logit, support_y))
                support_pred = pred_func(support_logit)
                support_accuracy = accuracy_func(support_pred, support_y)
                support_auroc = auroc_func(support_pred, support_y)
                self.support_accuracy(support_accuracy)
                self.support_accuracy_list(support_accuracy)
                self.support_auroc(support_auroc)
                self.support_auroc_list(support_auroc)

            query_logit = net(query_x)
            query_losses.append(loss_func(query_logit, query_y))
            with torch.no_grad():
                query_pred = pred_func(query_logit)
                query_accuracy = accuracy_func(query_pred, query_y)
                query_auroc = auroc_func(query_pred, query_y)
                self.query_accuracy(query_accuracy)
                self.query_accuracy_list(query_accuracy)
                self.query_auroc(query_auroc)
                self.query_auroc_list(query_auroc)

            torchopt.recover_state_dict(self.backbone, backbone_state_dict)
            self.logged_values.append(
                {
                    "support_loss": support_losses[-1].item(),
                    "query_loss": query_losses[-1].item(),
                    "support_accuracy": support_accuracy.item(),
                    "query_accuracy": query_accuracy.item(),
                    "support_auroc": support_auroc.item(),
                    "query_auroc": query_auroc.item(),
                }
            )

    def get_head(self, task: Task) -> torch.nn.Module:
        return new_classification_head(
            task.task_target,
            task.classes,
            self.num_backbone_features,
            task.support[0].device,
        )

    def get_loss_func(
        self, task: Task
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        return get_loss_func(task.task_target, task.classes, task.support[0].device)

    def get_pred_func(self, task: Task) -> Callable[[torch.Tensor], torch.Tensor]:
        match task:
            case Task(task_target=TaskTarget.MULTICLASS_CLASSIFICATION):
                return partial(torch.softmax, dim=-1)
            case Task(
                task_target=(
                    TaskTarget.MULTILABEL_CLASSIFICATION
                    | TaskTarget.BINARY_CLASSIFICATION
                )
            ):
                return torch.sigmoid
            case Task(task_target=(TaskTarget.ORDINAL_REGRESSION)):
                return partial(torch.softmax, dim=-1)
            case Task(task_target):
                raise NotImplementedError(
                    f"Task target {task_target} not yet implemented"
                )

    def get_metric_func(
        self, task: Task
    ) -> tuple[
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    ]:
        return (
            get_accuracy_func(task.task_target, task.classes, task.support[0].device),
            get_auroc_func(task.task_target, task.classes, task.support[0].device),
        )



import logging
from collections import OrderedDict
from functools import partial
from typing import Callable, Any

import lightning.pytorch as pl
import torch
import torchmetrics
from omegaconf import DictConfig
from torch.optim import Optimizer

from torchcross.cd.heads import new_classification_head
from torchcross.cd.losses import get_loss_func
from torchcross.cd.metrics import get_accuracy_func, get_auroc_func
from torchcross.data.task import TaskTarget, TaskDescription
from torchcross.models.lightning import Classifier


class CrossDomainClassifier(Classifier):
    def __init__(
        self,
        backbone: tuple[torch.nn.Module, int],
        optimizer: Callable[[...], Optimizer],
        hparams: DictConfig,
        *args,
        **kwargs,
    ) -> None:
        super(Classifier, self).__init__(*args, **kwargs)
        self.save_hyperparameters(hparams)

        self.backbone, self.num_backbone_features = backbone
        self.optimizer = optimizer

        self.train_accuracy = torchmetrics.MeanMetric()
        self.train_AUROC = torchmetrics.MeanMetric()
        self.val_accuracy = torchmetrics.MeanMetric()
        self.val_AUROC = torchmetrics.MeanMetric()
        self.test_accuracy = torchmetrics.MeanMetric()
        self.test_AUROC = torchmetrics.MeanMetric()

        self.heads = torch.nn.ModuleDict()

    def forward(
        self,
        x: torch.Tensor,
        task_description: TaskDescription,
        create_head: bool = False,
    ) -> torch.Tensor:
        if x.shape[1] == 1:
            x = x.repeat(1, 3, 1, 1)
        head = self.get_head(task_description, create=create_head)
        net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))
        return net(x)

    def training_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        (x, y), batch_task_description = batch

        if x.shape[1] == 1:
            x = x.repeat(1, 3, 1, 1)

        head = self.get_head(batch_task_description, create=True)
        loss_func = self.get_loss_func(batch_task_description)
        pred_func = self.get_pred_func(batch_task_description)
        accuracy_func, auroc_func = self.get_metric_funcs(batch_task_description)
        net = torch.nn.Sequential(OrderedDict(backbone=self.backbone, head=head))

        logits = net(x)
        with torch.no_grad():
            pred = pred_func(logits)
            self.train_accuracy(accuracy_func(pred, y))
            self.train_AUROC(auroc_func(pred, y))
        loss = loss_func(logits, y)

        # self.log("loss/train", loss, batch_size=len(y))
        self.log_dict(
            {
                "loss/train": loss,
                "accuracy/train": self.train_accuracy,
                "AUROC/train": self.train_AUROC,
            },
            batch_size=len(y),
            prog_bar=True,
        )
        return loss

    def validation_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        if len(self.heads) == 0:
            logging.info("No heads found. Skipping validation step.")
            return
        loss, accuracy, auroc, batch_size = self._shared_eval_step(batch)
        self.val_accuracy(accuracy)
        self.val_AUROC(auroc)

        self.log_dict(
            {
                "loss/val": loss,
                "accuracy/val": self.val_accuracy,
                "AUROC/val": self.val_AUROC,
            },
            batch_size=batch_size,
            prog_bar=True,
        )

    def test_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        loss, accuracy, auroc, batch_size = self._shared_eval_step(batch)
        self.test_accuracy(accuracy)
        self.test_AUROC(auroc)

        self.log_dict(
            {
                "loss/test": loss,
                "accuracy/test": self.test_accuracy,
                "AUROC/test": self.test_AUROC,
            },
            batch_size=batch_size,
            prog_bar=True,
        )

    def _shared_eval_step(self, batch):
        (x, y), batch_task_description = batch
        loss_func = self.get_loss_func(batch_task_description)
        pred_func = self.get_pred_func(batch_task_description)
        accuracy_func, auroc_func = self.get_metric_funcs(batch_task_description)
        logits = self.forward(x, batch_task_description)
        pred = pred_func(logits)
        accuracy = accuracy_func(pred, y)
        auroc = auroc_func(pred, y)
        loss = loss_func(logits, y)
        return loss, accuracy, auroc, len(y)

    def predict_step(
        self,
        batch: tuple[torch.Tensor, TaskDescription],
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> torch.Tensor:
        x, batch_task_description = batch
        pred_func = self.get_pred_func(batch_task_description)
        pred_to_label_func = self.get_pred_to_label_func(batch_task_description)

        logits = self.forward(x, batch_task_description)
        pred = pred_func(logits)
        return pred_to_label_func(pred)

    def get_head(
        self, task_description: TaskDescription, create=False
    ) -> torch.nn.Module:
        task_identifier = task_description.task_identifier
        if task_identifier not in self.heads:
            if create:
                new_head = self._get_head(task_description)
                self.heads[task_identifier] = new_head
                self.optimizers().add_param_group({"params": new_head.parameters()})
            else:
                raise ValueError(f"Head for task {task_identifier} not found.")
        return self.heads[task_identifier]

    def _get_head(self, task_description: TaskDescription) -> torch.nn.Module:
        return new_classification_head(
            task_description.task_target,
            task_description.classes,
            self.num_backbone_features,
            self.device,
        )

    def get_loss_func(
        self, task_description: TaskDescription
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        return get_loss_func(
            task_description.task_target, task_description.classes, self.device
        )

    def get_pred_func(
        self, task_description: TaskDescription
    ) -> Callable[[torch.Tensor], torch.Tensor]:
        match task_description:
            case TaskDescription(task_target=TaskTarget.MULTICLASS_CLASSIFICATION):
                return partial(torch.softmax, dim=-1)
            case TaskDescription(
                task_target=(
                    TaskTarget.MULTILABEL_CLASSIFICATION
                    | TaskTarget.BINARY_CLASSIFICATION
                )
            ):
                return torch.sigmoid
            case TaskDescription(task_target=(TaskTarget.ORDINAL_REGRESSION)):
                return partial(torch.softmax, dim=-1)
            case TaskDescription(task_target):
                raise NotImplementedError(
                    f"Task target {task_target} not yet implemented"
                )

    def get_pred_to_label_func(
        self, task_description: TaskDescription
    ) -> Callable[[torch.Tensor], torch.Tensor]:
        match task_description:
            case TaskDescription(task_target=TaskTarget.MULTICLASS_CLASSIFICATION):
                return lambda x: x.argmax(dim=1)
            case TaskDescription(
                task_target=(
                    TaskTarget.MULTILABEL_CLASSIFICATION
                    | TaskTarget.BINARY_CLASSIFICATION
                )
            ):
                return lambda x: x > 0.5
            case TaskDescription(task_target=(TaskTarget.ORDINAL_REGRESSION)):
                class_keys = torch.tensor(
                    task_description.classes.keys(),
                    dtype=torch.long,
                    device=self.device,
                )
                return lambda x: class_keys[x.argmax(dim=1)]
            case TaskDescription(task_target):
                raise NotImplementedError(
                    f"Task target {task_target} not yet implemented"
                )

    def get_metric_funcs(
        self, task_description: TaskDescription
    ) -> tuple[
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    ]:
        return (
            get_accuracy_func(
                task_description.task_target, task_description.classes, self.device
            ),
            get_auroc_func(
                task_description.task_target, task_description.classes, self.device
            ),
        )

    def configure_optimizers(
        self,
    ) -> Optimizer | tuple[list[Optimizer], list[Any]]:
        optimizer = self.optimizer(params=self.parameters())

        if "lr_scheduler" in self.hparams:
            scheduler = self.hparams.lr_scheduler(optimizer)
            return [optimizer], [scheduler]

        return optimizer



from typing import Any, Sequence, Tuple, Union

import torchopt
import hydra
import lightning.pytorch as pl
import torch
import torchmetrics
from omegaconf import DictConfig
from torch.optim import Optimizer
import torch.nn.functional as F
from torchmetrics.metric import Metric

from torchcross.data.task import Task


class MAML(pl.LightningModule):
    def __init__(
        self,
        net: torch.nn.Module,
        optim_cfg,
        hparams: DictConfig,
        num_classes,
        inductive=True,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.cfg = hparams
        self.save_hyperparameters(hparams)

        self.net = net

        self.optim_cfg = optim_cfg

        self.inductive = inductive

        self.automatic_optimization = False

        self.train_support_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.train_query_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.val_support_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.val_query_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.test_support_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )
        self.test_query_accuracy = torchmetrics.Accuracy(
            "multiclass", num_classes=num_classes
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)

    def outer_loop(
        self, batch: Sequence[Task], train: bool, metrics: dict[str, Metric]
    ):
        inner_optimizer = self.optim_cfg.inner_optimizer(self.net)
        # inner_optimizer = torchopt.MetaSGD(self.net, lr=1e-1)

        mode = "metatrain" if train else "metaval"

        support_losses = []
        query_losses = []

        net_state_dict = torchopt.extract_state_dict(
            self.net, by="reference", detach_buffers=True
        )
        optim_state_dict = torchopt.extract_state_dict(inner_optimizer, by="reference")

        for task in batch:
            support_x, support_y = task.support
            query_x, query_y = task.query

            if support_x.shape[1] == 1:
                support_x = support_x.repeat(1, 3, 1, 1)
                query_x = query_x.repeat(1, 3, 1, 1)

            self.net.train()

            for k in range(self.cfg.num_inner_steps):
                support_logit = self.net(support_x)
                loss = F.cross_entropy(support_logit, support_y)
                inner_optimizer.step(loss)

            if self.inductive:
                self.net.eval()

            with torch.no_grad():
                support_logit = self.net(support_x)
                support_losses.append(F.cross_entropy(support_logit, support_y))
                support_pred = torch.softmax(support_logit, dim=-1)
                metrics["support_accuracy"](support_pred, support_y)

            query_logit = self.net(query_x)
            query_losses.append(F.cross_entropy(query_logit, query_y))
            with torch.no_grad():
                query_pred = torch.softmax(query_logit, dim=-1)
                metrics["query_accuracy"](query_pred, query_y)

            torchopt.recover_state_dict(self.net, net_state_dict)
            torchopt.recover_state_dict(inner_optimizer, optim_state_dict)

        support_loss = sum(support_losses) / len(support_losses)
        query_loss = sum(query_losses) / len(query_losses)
        self.log(f"loss/{mode}/support", support_loss, batch_size=len(batch))
        self.log(f"loss/{mode}/query", query_loss, batch_size=len(batch), prog_bar=True)

        return query_loss

    def training_step(self, batch: Any, batch_idx: int):
        outer_optimizer = self.optimizers()
        outer_optimizer.zero_grad()

        loss = self.outer_loop(
            batch,
            True,
            {
                "support_accuracy": self.train_support_accuracy,
                "query_accuracy": self.train_query_accuracy,
            },
        )

        self.manual_backward(loss)

        # for pn, pp in self.named_parameters():
        #     self.logger.experiment.add_histogram(
        #         "parameters/" + pn, pp, self.global_step
        #     )
        # for pn, pp in self.named_parameters():
        #     self.logger.experiment.add_histogram(
        #         "gradients/" + pn, pp.grad, self.global_step
        #     )

        self.log(
            "accuracy/metatrain/support_accuracy",
            self.train_support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metatrain/query_accuracy",
            self.train_query_accuracy,
            prog_bar=True,
        )

        self.optimizers().step()

    def validation_step(self, batch: Any, batch_idx: int):
        torch.set_grad_enabled(True)
        self.train()
        self.outer_loop(
            batch,
            False,
            {
                "support_accuracy": self.val_support_accuracy,
                "query_accuracy": self.val_query_accuracy,
            },
        )
        torch.set_grad_enabled(False)

        self.log(
            "accuracy/metaval/support_accuracy",
            self.val_support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metaval/query_accuracy",
            self.val_query_accuracy,
            prog_bar=True,
        )

    def test_step(self, batch: Any, batch_idx: int):
        torch.set_grad_enabled(True)
        self.train()
        self.outer_loop(
            batch,
            False,
            {
                "support_accuracy": self.test_support_accuracy,
                "query_accuracy": self.test_query_accuracy,
            },
        )
        torch.set_grad_enabled(False)

        self.log(
            "accuracy/metaval/support_accuracy",
            self.test_support_accuracy,
            prog_bar=False,
        )
        self.log(
            "accuracy/metaval/query_accuracy",
            self.test_query_accuracy,
            prog_bar=True,
        )

    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Tuple[Sequence[Optimizer], Sequence[Any]]]:

        outer_optimizer = self.optim_cfg.outer_optimizer(params=self.parameters())

        if self.optim_cfg.use_lr_scheduler:
            scheduler = hydra.utils.instantiate(
                self.optim_cfg.lr_scheduler, optimizer=outer_optimizer
            )
            return [outer_optimizer], [scheduler]

        return outer_optimizer



import random
from abc import ABC
import bisect
from collections.abc import Sequence

import numpy as np
import torch
from torch.utils.data import (
    Dataset,
    ConcatDataset,
    BatchSampler,
    RandomSampler,
    SequentialSampler,
    IterableDataset,
)

from torchcross.utils.collate_fn import identity
from .task import TaskTarget, TaskDescription


class TaskSource(Dataset, ABC):
    task_target: TaskTarget
    labels: np.ndarray | torch.Tensor | Sequence
    classes: dict[int, str] = None
    task_identifier: str = ""


class BatchedTaskSource(TaskSource, IterableDataset):
    def __init__(
        self,
        task_source: TaskSource,
        batch_size: int,
        shuffle: bool = True,
        drop_last: bool = False,
        collate_fn=None,
    ):
        self.task_source = task_source
        self.task_target = task_source.task_target
        self.classes = task_source.classes
        self.task_identifier = task_source.task_identifier

        sampler = (
            RandomSampler(self.task_source)
            if shuffle
            else SequentialSampler(self.task_source)
        )
        self.batch_sampler = BatchSampler(sampler, batch_size, drop_last)
        self.collate_fn = collate_fn if collate_fn else identity

    def __iter__(self):
        for batch in self.batch_sampler:
            yield self.collate_fn([self.task_source[i] for i in batch])

    def __len__(self):
        return len(self.batch_sampler)


class RandomChainTaskSource(TaskSource, IterableDataset):
    def __init__(self, task_sources: list[BatchedTaskSource]):
        self.task_sources = task_sources

    def __iter__(self):
        iterators = [iter(ts) for ts in self.task_sources]
        # Calculate the weights of each iterator based on the batched task source length
        weights = [len(ts) for ts in self.task_sources]
        task_descriptions = [
            TaskDescription(
                ts.task_target, ts.classes, ts.task_identifier
            ) for ts in self.task_sources
        ]
        while iterators:
            # Randomly select an iterator with a probability proportional to its weight
            i = random.choices(range(len(self.task_sources)), weights=weights)[0]
            selected_iterator = iterators[i]
            try:
                # Return one element from the selected iterator
                yield next(selected_iterator), task_descriptions[i]
            except StopIteration:
                # The iterator is exhausted, so remove it from the list of iterators
                iterators.remove(selected_iterator)

    def __len__(self):
        return sum(len(ts) for ts in self.task_sources)


class ConcatTaskSource(TaskSource, ConcatDataset):
    datasets: list[TaskSource]

    def __getitem__(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length"
                )
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return (
            self.datasets[dataset_idx][sample_idx],
            TaskDescription(
                self.datasets[dataset_idx].task_target,
                self.datasets[dataset_idx].classes,
                self.datasets[dataset_idx].task_identifier,
            ),
        )
